#!/usr/local/bin/perl
#
#   startproxy.cgi--  Part of CGIProxy 1.2:  Initiate a browsing session
#     through nph-proxy.cgi.
#
#   Copyright (C) 1996, 1998-1999 by James Marshall, james@jmarshall.com
#
#   For the latest, see http://www.jmarshall.com/tools/cgiproxy/
#
#   This script initiates a browsing session through nph-proxy.cgi, prompting
#     the user with a form if needed.  Generates an appropriate "Location:"
#     response with the target URL correctly encoded in PATH_INFO.
#   Efficiency isn't as critical here as in the main nph-proxy.cgi script,
#     since this is normally called once per browsing session.  Therefore,
#     as much as possible is offloaded out of nph-proxy.cgi into here.  In
#     fact, this script contains way more features than it really needs; 
#     the basic idea is trivial.
#
#
#   CONFIGURATION:
#     None, unless these defaults in nph-proxy.cgi are changed:  
#       . If the name of nph-proxy.cgi is changed, set $proxyname[s] below.
#       . Make sure &proxy_encode() matches the one in nph-proxy.cgi.
#

#------ user configuration ------------------------------------------------

# If you change the name of the main proxy script, reflect it here.
# Changed to support SSL double-proxy.  These two end up in the Location:
#   response header, so they can be relative or absolute URLs.
$proxyname=     'http://.../' ;
$proxyname_ssl= 'https://.../' ;


# To make the above lines consistent with nph- scripts, remove trailing 
#   slashes.
$proxyname=~     s#/$## ;
$proxyname_ssl=~ s#/$## ;


# Create your own proxy_encode() and proxy_decode() to tranform the target
#   URL to and from the format that will be stored in PATH_INFO.  The encoded
#   form should only contain characters that are legal in PATH_INFO, i.e.
#   most printable chars except ? and #, and no two adjacent slashes ("//").
#   Try not to make the resulting string too long, either.
# This must match what is in nph-proxy.cgi, or whatever you called it.

sub proxy_encode {
    local($URL)= @_ ;
    $URL=~ s#^([\w+.-]+)://#$1/# ;                 # http://xxx -> http/xxx
#    $URL=~ s/(.)/ sprintf('%02x',ord($1)) /ge ;   # each char -> 2-hex
#    $URL=~ tr/a-zA-Z/n-za-mN-ZA-M/ ;              # rot-13
    return $URL ;
}

#------ end of (useful) user configuration --------------------------------


$PROXY_VERSION= '1.2' ;

# Read the URL from the query input
($URL)= ($ENV{'QUERY_STRING'}=~ /^URL=([^&]*)/) ;

# un-encode the URL
$URL=~ s/\+/ /g ;
$URL=~ s/%([\da-fA-F]{2})/pack("c", hex($1))/ge ;

$URL=~ s/^\s+|\s+$//g ;    # strip leading or trailing spaces

# Show start form if URL is empty (which may be caused by wrong QUERY_STRING).
&showstartform unless $URL ne '' ;

# scheme defaults to HTTP
$URL= "http://$URL" unless $URL=~ m#^[\w+.-]+://# ;

# Parse URL
($scheme, $authority, $path)= 
    ($URL=~ m#^([\w+.-]+)://([^/?]*)(.*)$#i) ;
$scheme=~ tr/A-Z/a-z/ ;
$path= "/$path" if $path!~ m#^/# ;     # if path is '' or contains only query


if ($scheme eq 'http' or $scheme eq 'https') {

    ($host,$portst)= $authority=~ /^([^:]*)(:?.*)$/ ;

    # If $host is long integer like 3467251275, break it into a.b.c.d format.
    # This is for big-endian; reverse the list for little-endian.
    $host= join('.', $host>>24 & 255, $host>>16 & 255, $host>>8 & 255, 
		     $host & 255)
	if $host=~ /^\d+$/ ;

    # Allow abbreviated host names if host doesn't exist on LAN
    $host= "www.$host.com"      unless ($host=~ /\./) || gethostbyname($host) ;

    $authority= "$host$portst" ;


} elsif ($scheme eq 'ftp') {

    # Don't mess with it if there's username/password embedded (which you
    #   REALLY shouldn't do).
    unless ($authority=~ /@/) {
	($host, $portst)= $authority=~ /^([^:]*)(:?.*)$/ ;

	# Allow abbreviated host names if host doesn't exist on LAN
	$host= "ftp.$host.com"  unless ($host=~ /\./) || gethostbyname($host) ;

	$authority= "$host$portst" ;
    }


} else {

    # Warn the user against unsupported schemes
    &HTMLdie("Sorry, only HTTP and FTP browsing are currently supported.") ;
}


$URL= &proxy_encode("$scheme://$authority$path") ;


# Print the Location: header
print $scheme eq 'https'
     ? "Location: $proxyname_ssl/$URL\n\n"
     : "Location: $proxyname/$URL\n\n" ;

exit ;



# Present entry form
sub showstartform {
    print <<EOF ;
Content-type: text/html

<html>
<head>
<title>Start Using CGI Proxy</title>
</head>
<body>

<h1>CGI Proxy</h1>

<p>Start browsing through this CGI-based proxy by entering a URL below.
Only HTTP, HTTPS, and FTP URLs are supported.  Not all functions will work 
(e.g. HTTP Basic authentication, or some JavaScript), but most pages will 
be fine.

<form action="$ENV{'SCRIPT_NAME'}" method=get>
<input name="URL" size=50>
<p><input type=submit value="   Begin browsing   ">
</form>
<p>
<hr>
<a href="http://www.jmarshall.com/tools/cgiproxy/"><i>CGIProxy $PROXY_VERSION</i></a>
<p>
</body>
</html>
EOF

    exit ;
}


# Die, outputting HTML error page
sub HTMLdie {
    local($msg)= @_ ;
    print <<EOF ;
Content-type: text/html

<html>
<head>
<title>CGI Proxy Error</title>
</head>
<body>
<h1>CGI Proxy Error</h1>
<h3>$msg</h3>
<p>
<hr>
<a href="http://www.jmarshall.com/tools/cgiproxy/"><i>CGIProxy $PROXY_VERSION</i></a>
<p>
</body>
</html>
EOF

    exit ;
}
